#include <termio.h>
#include <cstdlib>
#include <unistd.h>

using namespace std;

#include "video.h"

extern int Quittah;

static const bool AnsiOpt             = true;	// Be not brainless?
static const bool Colors              = true;	// Use colors?
static const bool Trust_TabSize       = true;	// Assume regular tabs of 8?
static const bool Use_VerticalTabs    = true;	// Enable VT?
static const bool Trust_DefaultColors = false;	// Trust white on black?

// These aren't ansi and so won't
// pass -ansi without being declared.
extern "C" int ioctl(int,int,...);
extern "C" void usleep(unsigned long);

int COLS=80, LINES=25;

Video video;

void Video::ShowCurs(const bool c)
{
	cout << (c ? "\33[?25h" : "\33[?25l");
}

void Video::GetScreenGeometry()
{
#if defined(DJGPP) || defined(__BORLANDC__)
    struct text_info w;
    gettextinfo(&w);
    LINES=w.screenheight;
    COLS =w.screenwidth;
#else
#ifdef TIOCGWINSZ
    struct winsize w;
    if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &w) >= 0)
    {
        LINES=w.ws_row;
        COLS =w.ws_col;
    }
#else
#ifdef WIOCGETD
    struct uwdata w;
    if(ioctl(STDOUT_FILENO, WIOCGETD, &w) >= 0)
    {
        LINES = w.uw_height / w.uw_vs;
        COLS  = w.uw_width / w.uw_hs;
    }
#endif
#endif
#endif
}

void Video::SetAttr(unsigned char n)
{
	if(n==TextAttr)return;
    if(Colors)
    {
    	cout << "\33[";
    	if(n != 7 || !Trust_DefaultColors)
    	{
   			static const char Swap[] = "04261537";
	    	if(AnsiOpt)
	    	{
	    		bool pp=false, r=false;
    	
		    	if((TextAttr&0x80) > (n&0x80)
    			|| (TextAttr&0x08) > (n&0x08))
		    	{
		    		cout << '0'; pp = true;
		    		TextAttr = 0x07;
		    		r = !Trust_DefaultColors;
	    		}
				
		    	if((n&0x08) && !(TextAttr&0x08)){if(pp)cout<<';';cout<<'1';pp=true;}
		    	if((n&0x80) && !(TextAttr&0x80)){if(pp)cout<<';';cout<<'5';pp=true;}
	       		
		      	if((n&7) != (TextAttr&7) || r)
		      	{
		      		if(pp)cout << ';';
		      		cout << '3' << Swap[n&7];
		      		pp = true;
	    	   	}
		       	if((n&0x70) != (TextAttr&0x70) || r)
		       	{
		       		if(pp)cout << ';';
		       		cout << '4' << Swap[(n>>4)&7];
	       		}
	       	}
   	 	   	else
  		 	{
  		 	   	cout << '0';
  		 	   	if(n&0x08)cout << ";1";
				if(n&0x80)cout << ";5";
				cout << ";4" << Swap[(n>>4)&7]
					 << ";3" << Swap[n&7];
			}
		}
		cout << 'm';
   	}   	
   	TextAttr = n;
}

Video::Video()
{
	ioctl(0, TCGETA, &back);
	GetScreenGeometry();
	SetAttr(7);
	clear();
	cout << "\33(U";
}
Video::~Video()
{
	ioctl(0, TCSETA, &back);	/* restore original terminal mode	*/
	SetAttr(7);
	Goto(LINES, 1);
	cout << "\n\n\33(B\33[?2c";
	ShowCurs(1);
}

void Video::clear()
{
	cout << "\33[2J\33[H";
	wx=wy=1;
}

const unsigned char Video::getkey(int wait)
{
	struct termio term = back;
	term.c_lflag &= ~ECHO & ~ICANON;
	term.c_cc[VMIN] = wait;		/* when set to 0, getch() will return	*/
	ioctl(0, TCSETA, &term);	/* immediately, if no key was pressed	*/
								/* (return value is then EOF)		*/
	int c = getchar();
	if(c!=EOF||Quittah)return(unsigned char)c;
	Flush();
	usleep(20000);
	return 0;
}
Video &Video::Goto(int y,int x,const int v)
{
	if(v>=0)ShowCurs(v);
	
//	if(x==wx && y==wy)goto Done;
//	cout << "\33[" << (wy=y) << ';' << (wx=x) << 'H';

	register int oldx=wx, oldy=wy;
	if(x==oldx && y==oldy)goto Done;

	wx=x, wy=y;
	
	x--, y--;
	oldx--, oldy--;
	
	if(AnsiOpt)
	{
		if(Trust_TabSize)
		{
			if(x==(oldx&~7)+8)
			{
				cout << '\t';
				goto FixUp;
			}
			if(x==(oldx&~7)+16)
			{
				cout << "\t\t";
				goto FixUp;
			}
		}
		if(!x)
		{
			if(y==oldy+4) { cout << "\n\n\n\n"; goto Done; }
			if(y==oldy+3) { cout << "\n\n\n"; goto Done; }
			if(y==oldy+2) { cout << "\n\n"; goto Done; }
			if(y==oldy+1) { cout << '\n'; goto Done; }
			if(!y) { cout << "\33[H"; goto Done; }
			if(oldx)cout << '\r';
			goto FixUp;
		}
		if(x==oldx)goto FixUp;
		if(x==oldx-1)
		{
			cout << '\b';
		FixUp:	 if(y == oldy-1)cout << "\33[A";
			else if(Use_VerticalTabs && y == oldy+1)cout << '\v';
			else if(Use_VerticalTabs && y == oldy+2)cout << "\v\v";
			else if(Use_VerticalTabs && y == oldy+3)cout << "\v\v\v";
			else if(y < oldy)cout << "\33[" << oldy-y << 'A';
			else if(y > oldy)cout << "\33[" << y-oldy << 'B';
			goto Done;
		}
		if(y == oldy)
		{
			/***/if(x == oldx+1)cout << "\33[C";
			else if(x == oldx-2)cout << "\b\b";
			else if(x == oldx-3)cout << "\b\b\b";
			else if(x < oldx)cout << "\33[" << oldx-x << 'D';
			else if(x > oldx)cout << "\33[" << x-oldx << 'C';
			goto Done;
		}
	}
	cout << "\33[" << wy << ';' << wx << 'H';
	
Done:
	Flush();
	return *this;
}
